/*
 * Copyright 2009-2010 Maarten Baert
 * maarten-baert@hotmail.com
 * http://www.maartenbaert.be/
 * 
 * This file is part of ExtremePhysics.
 * 
 * ExtremePhysics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * ExtremePhysics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with ExtremePhysics. If not, see <http://www.gnu.org/licenses/>.
 * 
 * File: ep_HingeJoint.cpp
 * Implementation of ep_HingeJoint.
 */

#include "ExtremePhysics.h"

void ep_HingeJoint::Init(ep_World* _world, ep_Body* _body1, ep_Body* _body2, double _x1, double _y1, double _x2, double _y2, double _referencerotation) {
	
	// add to linked list
	main = _world->main;
	world = _world;
	prev = world->last_hingejoint;
	next = NULL;
	if(prev==NULL) world->first_hingejoint = this;
	else prev->next = this;
	world->last_hingejoint = this;
	world->current_hingejoint = this;
	
	// initialize variables
	id = ++world->idcounter_hingejoints;
	
	body1 = _body1;
	body2 = _body2;
	
	x1 = _x1;
	y1 = _y1;
	x2 = _x2;
	y2 = _y2;
	referencerotation = _referencerotation;
	maxforce = 0.0;
	
	maxmotortorque = 0.0;
	motorvel = 0.0;
	
	limit1_maxtorque = 0.0;
	limit2_maxtorque = 0.0;
	limit1_rot = 0.0;
	limit2_rot = 0.0;
	
	xforce = 0.0;
	yforce = 0.0;
	torque = 0.0;
	limittorque = 0.0;
	
	// initialize links
	link1.hingejoint = this;
	link2.hingejoint = this;
	link1.other_body = body2;
	link2.other_body = body1;
	
	// add link1 to body1's joint list
	link1.prev = body1->last_hingejointlink;
	link1.next = NULL;
	if(link1.prev==NULL) body1->first_hingejointlink = &link1;
	else link1.prev->next = &link1;
	body1->last_hingejointlink = &link1;
	
	// add link2 to body1's joint list
	link2.prev = body2->last_hingejointlink;
	link2.next = NULL;
	if(link2.prev==NULL) body2->first_hingejointlink = &link2;
	else link2.prev->next = &link2;
	body2->last_hingejointlink = &link2;
	
	// awake connected bodies
	body1->Awake(true, false);
	body2->Awake(true, false);
	
	++world->hingejointcount;
	
}

void ep_HingeJoint::DeInit() {
	
	// remove from linked list
	if(prev==NULL) world->first_hingejoint = next;
	else prev->next = next;
	if(next==NULL) world->last_hingejoint = prev;
	else next->prev = prev;
	if(world->current_hingejoint==this) world->current_hingejoint = NULL;
	
	// remove link1 from body1's joint list
	if(link1.prev==NULL) body1->first_hingejointlink = link1.next;
	else link1.prev->next = link1.next;
	if(link1.next==NULL) body1->last_hingejointlink = link1.prev;
	else link1.next->prev = link1.prev;
	
	// remove link2 from body2's joint list
	if(link2.prev==NULL) body2->first_hingejointlink = link2.next;
	else link2.prev->next = link2.next;
	if(link2.next==NULL) body2->last_hingejointlink = link2.prev;
	else link2.next->prev = link2.prev;
	
	// awake connected bodies
	body1->Awake(true, false);
	body2->Awake(true, false);
	
	--world->hingejointcount;
	
}

void ep_HingeJoint::SetMaxForce(double _maxforce) {
	if(_maxforce<0.0) _maxforce = 0.0;
	if(maxforce!=_maxforce) {
		maxforce = _maxforce;
		body1->Awake(false, false);
		body2->Awake(false, false);
	}
#if EP_COMPILE_DEBUGMESSAGES
	main->Message(EP_MESSAGELEVEL_NORMALCHANGE, "Changed maximum force of hinge joint %lu in world %lu.", id, world->id);
#endif
}

void ep_HingeJoint::SetMotor(double _maxmotortorque, double _motorvel) {
	if(_maxmotortorque<0.0) _maxmotortorque = 0.0;
	if(maxmotortorque!=_maxmotortorque || motorvel!=_motorvel) {
		maxmotortorque = _maxmotortorque;
		motorvel = _motorvel;
		body1->Awake(false, false);
		body2->Awake(false, false);
	}
#if EP_COMPILE_DEBUGMESSAGES
	main->Message(EP_MESSAGELEVEL_NORMALCHANGE, "Changed motor settings of hinge joint %lu in world %lu.", id, world->id);
#endif
}

void ep_HingeJoint::SetLowerLimit(double maxlimittorque, double rotation) {
	if(maxlimittorque<0.0) maxlimittorque = 0.0;
	if(limit1_maxtorque!=maxlimittorque || limit1_rot!=rotation) {
		limit1_maxtorque = maxlimittorque;
		limit1_rot = rotation;
		body1->Awake(false, false);
		body2->Awake(false, false);
	}
#if EP_COMPILE_DEBUGMESSAGES
	main->Message(EP_MESSAGELEVEL_NORMALCHANGE, "Changed lower limit of hinge joint %lu in world %lu.", id, world->id);
#endif
}

void ep_HingeJoint::SetUpperLimit(double maxlimittorque, double rotation) {
	if(maxlimittorque<0.0) maxlimittorque = 0.0;
	if(limit2_maxtorque!=maxlimittorque || limit2_rot!=rotation) {
		limit2_maxtorque = maxlimittorque;
		limit2_rot = rotation;
		body1->Awake(false, false);
		body2->Awake(false, false);
	}
#if EP_COMPILE_DEBUGMESSAGES
	main->Message(EP_MESSAGELEVEL_NORMALCHANGE, "Changed upper limit of hinge joint %lu in world %lu.", id, world->id);
#endif
}

