/*
 * Copyright 2009-2010 Maarten Baert
 * maarten-baert@hotmail.com
 * http://www.maartenbaert.be/
 * 
 * This file is part of ExtremePhysics.
 * 
 * ExtremePhysics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * ExtremePhysics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with ExtremePhysics. If not, see <http://www.gnu.org/licenses/>.
 * 
 * File: ep_World.cpp
 * Implementation of ep_World.
 */

#include "ExtremePhysics.h"

void ep_World::Init(ep_Main* _main) {
	
	// add to linked list
	main = _main;
	prev = main->last_world;
	next = NULL;
	if(prev==NULL) main->first_world = this;
	else prev->next = this;
	main->last_world = this;
	main->current_world = this;
	
	// initialize variables
	id = ++main->idcounter_worlds;
	
	idcounter_polygons = 0;
	polygoncount = 0;
	first_polygon = NULL;
	last_polygon = NULL;
	current_polygon = NULL;
	
	idcounter_bodies = 0;
	bodycount = 0;
	first_body = NULL;
	last_body = NULL;
	current_body = NULL;
	
	idcounter_contacts = 0;
	contactcount = 0;
	first_contact = NULL;
	last_contact = NULL;
	current_contact = NULL;
	freecontactcount = 0;
	freecontacts = NULL;
	
	idcounter_hingejoints = 0;
	hingejointcount = 0;
	first_hingejoint = NULL;
	last_hingejoint = NULL;
	current_hingejoint = NULL;
	
	idcounter_views = 0;
	viewcount = 0;
	first_view = NULL;
	last_view = NULL;
	current_view = NULL;
	
	shapecount = 0;
	forcecount = 0;
	
	timestep = 1.0;
	iterations = 10;
	contact_threshold = 0.0;
	velocity_threshold = 0.0;
	baumgarte_factor = 0.1;
	mass_bias = 0.6;
	
	horizontal = true;
	
	enable_sleeping = false;
	time_stable = 0.0;
	time_outofview = 0.0;
	stable_maxvel = 0.0;
	stable_maxrotvel = 0.0;
	
}

void ep_World::DeInit() {
	
	// destroy everything (in reverse order)
	while(first_view!=NULL) DestroyView(first_view);
	while(first_hingejoint!=NULL) DestroyHingeJoint(first_hingejoint);
	while(first_contact!=NULL) DestroyContact(first_contact);
	while(first_body!=NULL) DestroyBody(first_body);
	while(first_polygon!=NULL) DestroyPolygon(first_polygon);
	
	// destroy free contacts
	while(freecontacts!=NULL) {
		ep_Contact *t = freecontacts->next;
		ep_free(freecontacts);
		freecontacts = t;
	}
	
	// remove from linked list
	if(prev==NULL) main->first_world = next;
	else prev->next = next;
	if(next==NULL) main->last_world = prev;
	else next->prev = prev;
	if(main->current_world==this) main->current_world = NULL;
	
}

void ep_World::Clear() {
	
	// destroy everything (in reverse order)
	while(first_view!=NULL) DestroyView(first_view);
	while(first_hingejoint!=NULL) DestroyHingeJoint(first_hingejoint);
	while(first_contact!=NULL) DestroyContact(first_contact);
	while(first_body!=NULL) DestroyBody(first_body);
	while(first_polygon!=NULL) DestroyPolygon(first_polygon);
	
	// destroy free contacts
	while(freecontacts!=NULL) {
		ep_Contact *t = freecontacts->next;
		ep_free(freecontacts);
		freecontacts = t;
	}
	freecontactcount = 0;
	
	// reset id counters
	idcounter_polygons = 0;
	idcounter_bodies = 0;
	idcounter_hingejoints = 0;
	idcounter_views = 0;
	
}

// contact

ep_Contact* ep_World::CreateContact(ep_Shape* shape1, ep_Shape* shape2) {
	ep_Contact *contact;
	
	// free contacts left?
	if(freecontactcount>0) {
		// pop one contact
		contact = freecontacts;
		freecontacts = freecontacts->next;
		--freecontactcount;
	} else {
		// allocate memory
		contact = (ep_Contact*) ep_malloc(sizeof(ep_Contact));
		if(contact==NULL) {
#if EP_COMPILE_DEBUGMESSAGES
			main->Message(EP_MESSAGELEVEL_ERROR, "Could not create contact in world %lu, memory allocation failed.", id);
#endif
			return NULL;
		}
	}
	
	contact->Init(this, shape1, shape2);
	
	return contact;
}

void ep_World::DestroyContact(ep_Contact* contact) {
	
	contact->DeInit();
	
	// store free contact?
	if(freecontactcount<EP_MAXFREECONTACTS) {
		// push one contact
		contact->next = freecontacts; // 'next' is the only variable used here 
		freecontacts = contact;
		++freecontactcount;
	} else {
		// free the memory
		ep_free(contact);
	}
	
}

ep_Contact* ep_World::FindContact(unsigned long _id) {
	if(current_contact!=NULL) {
		if(current_contact->id==_id) return current_contact;
	}
	for(current_contact = first_contact; current_contact!=NULL; current_contact = current_contact->next) {
		if(current_contact->id==_id) return current_contact;
	}
	return NULL;
}

// polygon

ep_Polygon* ep_World::CreatePolygon(unsigned long vertexcount, unsigned long userdatasize) {
	ep_Polygon *polygon;
	
#if EP_COMPILE_DEBUGCHECKS
	// first some additional error checking
	// is the vertex count smaller than 3?
	if(vertexcount<3) {
#if EP_COMPILE_DEBUGMESSAGES
		main->Message(EP_MESSAGELEVEL_ERROR, "Could not create polygon in world %lu, vertex count is smaller than 3.", id);
#endif
		return NULL;
	}
#endif
	
	// allocate memory
	polygon = (ep_Polygon*) ep_malloc(sizeof(ep_Polygon)+vertexcount*sizeof(ep_PolygonVertex)+userdatasize);
	if(polygon==NULL) {
#if EP_COMPILE_DEBUGMESSAGES
		main->Message(EP_MESSAGELEVEL_ERROR, "Could not create polygon in world %lu, memory allocation failed.", id);
#endif
		return NULL;
	}
	
	polygon->Init(this, vertexcount);
	
#if EP_COMPILE_DEBUGMESSAGES
	main->Message(EP_MESSAGELEVEL_IMPORTANTCHANGE, "Created polygon %lu in world %lu.", polygon->id, id);
#endif
	
	return polygon;
}

bool ep_World::DestroyPolygon(ep_Polygon* polygon) {
	
#if EP_COMPILE_DEBUGCHECKS
	if(polygon->referencecount!=0) {
#if EP_COMPILE_DEBUGMESSAGES
		main->Message(EP_MESSAGELEVEL_ERROR, "Could not destroy polygon %lu in world %lu, polygon is still being used by %lu shapes.", polygon->id, id, polygon->referencecount);
#endif
		return false;
	}
#endif
	
	polygon->DeInit();
#if EP_COMPILE_DEBUGMESSAGES
	main->Message(EP_MESSAGELEVEL_IMPORTANTCHANGE, "Destroyed polygon %lu in world %lu.", polygon->id, id);
#endif
	ep_free(polygon);
	
	return true;
}

ep_Polygon* ep_World::FindPolygon(unsigned long _id) {
	if(current_polygon!=NULL) {
		if(current_polygon->id==_id) return current_polygon;
	}
	for(current_polygon = first_polygon; current_polygon!=NULL; current_polygon = current_polygon->next) {
		if(current_polygon->id==_id) return current_polygon;
	}
	return NULL;
}

// body

ep_Body* ep_World::CreateStaticBody(bool groundbody, unsigned long userdatasize) {
	ep_Body *body;
	
	// allocate memory
	body = (ep_Body*) ep_malloc(sizeof(ep_Body)+userdatasize);
	if(body==NULL) {
#if EP_COMPILE_DEBUGMESSAGES
		main->Message(EP_MESSAGELEVEL_ERROR, "Could not create body in world %lu, memory allocation failed.", id);
#endif
		return NULL;
	}
	
	body->Init(this, groundbody, true, true);
	
#if EP_COMPILE_DEBUGMESSAGES
	main->Message(EP_MESSAGELEVEL_IMPORTANTCHANGE, "Created body %lu in world %lu.", body->id, id);
#endif
	
	return body;
}

ep_Body* ep_World::CreateDynamicBody(bool norotation, unsigned long userdatasize) {
	ep_Body *body;
	
	// allocate memory
	body = (ep_Body*) ep_malloc(sizeof(ep_Body)+userdatasize);
	if(body==NULL) {
#if EP_COMPILE_DEBUGMESSAGES
		main->Message(EP_MESSAGELEVEL_ERROR, "Could not create body in world %lu, memory allocation failed.", id);
#endif
		return NULL;
	}
	
	body->Init(this, false, false, norotation);
	
#if EP_COMPILE_DEBUGMESSAGES
	main->Message(EP_MESSAGELEVEL_IMPORTANTCHANGE, "Created body %lu in world %lu.", body->id, id);
#endif
	
	return body;
}

bool ep_World::DestroyBody(ep_Body* body) {
	
	body->DeInit();
#if EP_COMPILE_DEBUGMESSAGES
	main->Message(EP_MESSAGELEVEL_IMPORTANTCHANGE, "Destroyed body %lu in world %lu.", body->id, id);
#endif
	ep_free(body);
	
	return true;
}

ep_Body* ep_World::FindBody(unsigned long _id) {
	if(current_body!=NULL) {
		if(current_body->id==_id) return current_body;
	}
	for(current_body = first_body; current_body!=NULL; current_body = current_body->next) {
		if(current_body->id==_id) return current_body;
	}
	return NULL;
}

// hinge joint

ep_HingeJoint* ep_World::CreateHingeJoint(ep_Body* body1, ep_Body* body2, double x1, double y1, double x2, double y2, double referencerotation, unsigned long userdatasize) {
	ep_HingeJoint *hingejoint;
	
#if EP_COMPILE_DEBUGCHECKS
	// first some additional error checking
	// do the bodies belong to this world?
	if(body1->world!=this) {
#if EP_COMPILE_DEBUGMESSAGES
		main->Message(EP_MESSAGELEVEL_ERROR, "Could not create hinge joint in world %lu, the first body does not belong to this world.", id);
#endif
		return NULL;
	}
	if(body2->world!=this) {
#if EP_COMPILE_DEBUGMESSAGES
		main->Message(EP_MESSAGELEVEL_ERROR, "Could not create hinge joint in world %lu, the second body does not belong to this world.", id);
#endif
		return NULL;
	}
	// are the bodies the same?
	if(body1==body2) {
#if EP_COMPILE_DEBUGMESSAGES
		main->Message(EP_MESSAGELEVEL_ERROR, "Could not create hinge joint in world %lu, body %lu can't be connected to itself.", id, body1->id);
#endif
		return NULL;
	}
	// are both bodies static?
	if(body1->isstatic && body2->isstatic) {
#if EP_COMPILE_DEBUGMESSAGES
		main->Message(EP_MESSAGELEVEL_ERROR, "Could not create hinge joint in world %lu, body %lu and body %lu are both static.", id, body1->id, body2->id);
#endif
		return NULL;
	}
#endif
	
	// allocate memory
	hingejoint = (ep_HingeJoint*) ep_malloc(sizeof(ep_HingeJoint)+userdatasize);
	if(hingejoint==NULL) {
#if EP_COMPILE_DEBUGMESSAGES
		main->Message(EP_MESSAGELEVEL_ERROR, "Could not create hinge joint in world %lu, memory allocation failed.", id);
#endif
		return NULL;
	}
	
	hingejoint->Init(this, body1, body2, x1, y1, x2, y2, referencerotation);
	
#if EP_COMPILE_DEBUGMESSAGES
	main->Message(EP_MESSAGELEVEL_IMPORTANTCHANGE, "Created hinge joint %lu in world %lu.", hingejoint->id, id);
#endif
	
	return hingejoint;
}

bool ep_World::DestroyHingeJoint(ep_HingeJoint* hingejoint) {
	
	hingejoint->DeInit();
#if EP_COMPILE_DEBUGMESSAGES
	main->Message(EP_MESSAGELEVEL_IMPORTANTCHANGE, "Destroyed hinge joint %lu in world %lu.", hingejoint->id, id);
#endif
	ep_free(hingejoint);
	
	return true;
}

ep_HingeJoint* ep_World::FindHingeJoint(unsigned long _id) {
	if(current_hingejoint!=NULL) {
		if(current_hingejoint->id==_id) return current_hingejoint;
	}
	for(current_hingejoint = first_hingejoint; current_hingejoint!=NULL; current_hingejoint = current_hingejoint->next) {
		if(current_hingejoint->id==_id) return current_hingejoint;
	}
	return NULL;
}

// view

ep_View* ep_World::CreateView(unsigned long userdatasize) {
	ep_View *view;
	
	// allocate memory
	view = (ep_View*) ep_malloc(sizeof(ep_View)+userdatasize);
	if(view==NULL) {
#if EP_COMPILE_DEBUGMESSAGES
		main->Message(EP_MESSAGELEVEL_ERROR, "Could not create view in world %lu, memory allocation failed.", id);
#endif
		return NULL;
	}
	
	view->Init(this);
	
#if EP_COMPILE_DEBUGMESSAGES
	main->Message(EP_MESSAGELEVEL_IMPORTANTCHANGE, "Created view %lu in world %lu.", view->id, id);
#endif
	
	return view;
}

bool ep_World::DestroyView(ep_View* view) {
	
	view->DeInit();
#if EP_COMPILE_DEBUGMESSAGES
	main->Message(EP_MESSAGELEVEL_IMPORTANTCHANGE, "Destroyed view %lu in world %lu.", view->id, id);
#endif
	ep_free(view);
	
	return true;
}

ep_View* ep_World::FindView(unsigned long _id) {
	if(current_view!=NULL) {
		if(current_view->id==_id) return current_view;
	}
	for(current_view = first_view; current_view!=NULL; current_view = current_view->next) {
		if(current_view->id==_id) return current_view;
	}
	return NULL;
}

// other

void ep_World::Awake() {
	// This function has the same effect as calling
	// Awake(false, true) for all bodies, but it's faster.
	for(ep_Body *body = first_body; body!=NULL; body = body->next) {
		body->stabletimer = 0.0;
	}
}

void ep_World::SetSettings(double _timestep, unsigned long _iterations, double _contact_threshold, double _velocity_threshold, double _baumgarte_factor, double _mass_bias) {
	_timestep = (_timestep<EP_EPSILON_MINTIMESTEP)? 0.0 : _timestep;
	_iterations = ep_max(1, _iterations);
	if(timestep!=_timestep || iterations!=_iterations || contact_threshold!=_contact_threshold || velocity_threshold!=_velocity_threshold || baumgarte_factor!=_baumgarte_factor || mass_bias!=_mass_bias) {
		timestep = _timestep;
		iterations = _iterations;
		contact_threshold = _contact_threshold;
		velocity_threshold = _velocity_threshold;
		baumgarte_factor = _baumgarte_factor;
		mass_bias = _mass_bias;
		if(timestep!=0.0) {
			for(ep_Body *body = first_body; body!=NULL; body = body->next) {
				body->damping_factor = pow(1.0-body->damping, timestep);
				body->rotdamping_factor = pow(1.0-body->rotdamping, timestep);
			}
		}
		Awake();
	}
#if EP_COMPILE_DEBUGMESSAGES
	main->Message(EP_MESSAGELEVEL_NORMALCHANGE, "Changed settings of world %lu.", id);
#endif
}

void ep_World::SetPrimaryAxis(bool _horizontal) {
	// probably the simplest function in ExtremePhysics
	// awake is not required here because this
	// setting doesn't make any real difference
	horizontal = _horizontal;
#if EP_COMPILE_DEBUGMESSAGES
	main->Message(EP_MESSAGELEVEL_NORMALCHANGE, "Changed settings of world %lu.", id);
#endif
}

void ep_World::SetSleeping(bool _enable_sleeping, double _time_stable, double _time_outofview, double _stable_maxvel, double _stable_maxrotvel) {
	if(enable_sleeping!= _enable_sleeping || time_stable!=_time_stable || time_outofview!=_time_outofview || stable_maxvel!=_stable_maxvel || stable_maxrotvel!=_stable_maxrotvel) {
		enable_sleeping = _enable_sleeping;
		time_stable = _time_stable;
		time_outofview = _time_outofview;
		stable_maxvel = _stable_maxvel;
		stable_maxrotvel = _stable_maxrotvel;
		if(!enable_sleeping) {
			for(ep_Body *body = first_body; body!=NULL; body = body->next) {
				body->issleeping = false;
				body->stabletimer = 0.0;
				body->outofviewtimer = 0.0;
			}
		}
	}
#if EP_COMPILE_DEBUGMESSAGES
	main->Message(EP_MESSAGELEVEL_NORMALCHANGE, "Changed sleeping settings of world %lu.", id);
#endif
}

bool ep_World::SimulateStep() {
	
	// contacts have to be updated before the simulation starts because of the sleeping system
	if(!UpdateContacts()) {
		return false;
	}
	if(enable_sleeping) {
		UpdateSleeping();
	}
	if(timestep!=0.0) {
		IntegrateVelocity();
		SolveVelocityConstraints();
		IntegratePosition();
	}
	
#if EP_COMPILE_DEBUGMESSAGES
	main->Message(EP_MESSAGELEVEL_NORMALCHANGE, "Simulated a step in world %lu.", id);
#endif
	return true;
}

