/*
 * Copyright 2009-2010 Maarten Baert
 * maarten-baert@hotmail.com
 * http://www.maartenbaert.be/
 * 
 * This file is part of ExtremePhysics.
 * 
 * ExtremePhysics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * ExtremePhysics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with ExtremePhysics. If not, see <http://www.gnu.org/licenses/>.
 * 
 * File: gm.cpp
 * Initialization of ep_Main and implementation of ep_Callback.
 */

#include "gm.h"

#define WIN32_LEAN_AND_MEAN
#define NOGDI
#include <windows.h>

#ifdef _DEBUG
class __fp_control87 {
	public:
	__fp_control87() {
		// disable floating point exceptions
		// (disables NaN, denormal, divide-by-zero, overflow, underflow and inexact)
		// This is used only for debugging. Other errors frequently cause instabilities and overflow errors,
		// disabling these errors will make debugging easier.
		// These exceptions are disabled by default in C++, but apparently they are enabled in GM.
		_control87(_EM_INVALID|_EM_DENORMAL|_EM_ZERODIVIDE|_EM_OVERFLOW|_EM_UNDERFLOW|_EM_INEXACT, _MCW_EM);
	}
} fp_control87;
#endif // _DEBUG

char *logfile = NULL;
bool showerrors = true;

__gmcallback gmcallback;
ep_Main epmain(&gmcallback);

void __gmcallback::DebugMessage(int level, const char* string) {
	if(logfile!=NULL) {
		FILE *f;
#ifdef _MSC_VER
		// VC++ complains about deprecation
		fopen_s(&f, logfile, "a");
#else
		f = fopen(logfile, "a");
#endif
		if(f!=NULL) {
			fprintf(f, "[%d] %s\n", level, string);
			fclose(f);
		}
	}
	if(level==1 && showerrors) {
		char messagebuffer[EP_MESSAGE_BUFFERSIZE+100+1];
#ifdef _MSC_VER
		// VC++ complains about deprecation
		_snprintf_s(messagebuffer, EP_MESSAGE_BUFFERSIZE+100+1, EP_MESSAGE_BUFFERSIZE+100, "[Level %d]\n\n%s\n\n(Press cancel to disable ExtremePhysics error messages)", level, string);
#else
		snprintf(messagebuffer, EP_MESSAGE_BUFFERSIZE, "[Level %d]\n\n%s\n\n(Press cancel to disable ExtremePhysics error messages)", level, string);
#endif
		messagebuffer[EP_MESSAGE_BUFFERSIZE+100] = '\0';
		if(MessageBoxA(NULL, messagebuffer, "ExtremePhysics error", MB_ICONERROR|MB_OKCANCEL)==IDCANCEL) {
			showerrors = false;
		}
	}
}

void __gmcallback::DrawBody(bool isstatic, bool issleeping, double x, double y) {
	GM_draw_set_color((isstatic)? d1 : d2);
	/*
	x = ep_round(x);
	y = ep_round(y);
	*/
	if(isstatic) {
		GM_draw_rectangle(x-4.0, y-4.0, x+4.0, y+4.0, GM_TRUE);
		GM_draw_line(x-4.0, y-4.0, x+4.0, y+4.0);
		GM_draw_line(x+4.0, y-4.0, x-4.0, y+4.0);
	} else {
		GM_draw_rectangle(x-3.0, y-3.0, x+3.0, y+3.0, GM_TRUE);
		if(issleeping) {
			GM_draw_rectangle(x-5.0, y-5.0, x+5.0, y+5.0, GM_TRUE);
		}
	}
}

void __gmcallback::DrawSegment(double x1, double y1, double x2, double y2) {
	GM_draw_line(x1, y1, x2, y2);
}

void __gmcallback::DrawCircle(double x, double y, double r) {
	GM_draw_circle(x, y, r, GM_TRUE);
}

void __gmcallback::DrawShapeLink(bool isstatic, double x1, double y1, double x2, double y2) {
	GM_draw_set_color(d1);
	GM_draw_line_width(x1, y1, x2, y2, (isstatic)? 0.4 : 2.0);
}

void __gmcallback::DrawContactLink(bool isstatic, double x1, double y1, double x2, double y2) {
	GM_draw_set_color(d2);
	GM_draw_line_width(x1, y1, x2, y2, (isstatic)? 0.4 : 2.0);
}

void __gmcallback::DrawJointLink(bool isstatic, double x1, double y1, double x2, double y2) {
	GM_draw_set_color(d3);
	GM_draw_line_width(x1, y1, x2, y2, (isstatic)? 0.4 : 2.0);
}

void __gmcallback::DrawView(double x1, double y1, double x2, double y2) {
	GM_draw_rectangle(x1, y1, x2, y2, GM_TRUE);
}

void __gmcallback::DrawVelocity(double x, double y, double xvel, double yvel, double rotvel) {
	if(xvel!=0.0 || yvel!=0.0) GM_draw_arrow(x, y, x+xvel*d1, y+yvel*d1, 5.0);
	if(rotvel!=0.0) GM_draw_circle(x, y, rotvel*d2, GM_TRUE);
}

void __gmcallback::DrawForce(double x, double y, double xforce, double yforce, double torque) {
	if(xforce!=0.0 || yforce!=0.0) GM_draw_arrow(x, y, x+xforce*d1, y+yforce*d1, 5.0);
	if(torque!=0.0) GM_draw_circle(x, y, torque*d2, GM_TRUE);
}

void __gmcallback::DrawContactPoint(double x1, double y1, double x2, double y2) {
	GM_draw_set_color(d1);
	GM_draw_line(x1, y1, x2, y2);
	GM_draw_rectangle(x1-1.5, y1-1.5, x1+1.5, y1+1.5, GM_FALSE);
	GM_draw_rectangle(x2-1.5, y2-1.5, x2+1.5, y2+1.5, GM_FALSE);
}

void __gmcallback::DrawJointPoint(double x1, double y1, double x2, double y2) {
	GM_draw_set_color(d2);
	GM_draw_line(x1, y1, x2, y2);
	GM_draw_rectangle(x1-1.5, y1-1.5, x1+1.5, y1+1.5, GM_TRUE);
	GM_draw_rectangle(x2-1.5, y2-1.5, x2+1.5, y2+1.5, GM_TRUE);
}

